class MBException(Exception):
    def __init__(self, message: str, errorcode: int):
        self.errorcode = errorcode
        message += f" Error code: {errorcode}."
        super().__init__(message)
